# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path

import pytest

import ansys.simai.core.errors as err
from ansys.simai.core import SimAIClient


def test_client_creation_invalid_path():
    with pytest.raises(err.ConfigurationNotFoundError):
        SimAIClient.from_config(path="/")


def test_client_creation_invalid_config():
    with pytest.raises(err.InvalidConfigurationError):
        SimAIClient.from_config(path=Path(__file__).resolve())


@pytest.mark.parametrize(
    "local_ver,latest_ver,expected",
    [
        ("1.1.0", "1.1.1", "available."),
        ("1.0.9-rc8", "1.0.9", "available."),
        ("1.0.9", "1.9.0", "required."),
    ],
)
def test_client_version_auto_warn(caplog, mocker, httpx_mock, local_ver, latest_ver, expected):
    """WHEN the SDK version is slightly outdated compared to what the API responds
    THEN a warning is printed
    """
    mocker.patch(
        "ansys.simai.core.client.__version__",
        local_ver,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://pypi.org/pypi/ansys-simai-core/json",
        json={"info": {"version": latest_ver}},
        status_code=200,
    )
    SimAIClient(
        url="https://test.test",
        organization="dummy",
        _disable_authentication=True,
        no_sse_connection=True,
        skip_version_check=False,
    )
    assert f"A new version of ansys-simai-core is {expected}" in caplog.text



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
.. _ref_model_recomputation:

Model recomputation
===================

This example demonstrates how to relaunch a model build using the latest
model configuration in a same project.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

simai_client = asc.from_config()

###############################################################################
# Get the project from the server
# -------------------------------

my_project = simai_client.projects.get(name="old-ps")

###############################################################################
# Get the model configuration
# ---------------------------
# Get the latest model configuration of the project.

last_build_config = my_project.last_model_configuration

###############################################################################
# Verify the project requirements
# -------------------------------
# Verify that the project meets the requirements for training (model building).

is_trainable_check = my_project.is_trainable()

###############################################################################
# If the project met the requirements, launch a model build.
# Otherwise, print the reasons the project does not meet the requirements.
if is_trainable_check:
    new_model = simai_client.models.build(last_build_config)
else:
    print(is_trainable_check.reason)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.simai.core.data.post_processings import (
    GlobalCoefficients,
    Slice,
    SurfaceEvolution,
    SurfaceVTP,
    SurfaceVTPTDLocation,
    VolumeVTU,
)
from ansys.simai.core.data.selection_post_processings import ExportablePPList, PPList
from ansys.simai.core.data.selections import Selection
from ansys.simai.core.errors import ApiClientError


@pytest.fixture()
def test_selection(
    geometry_factory,
    prediction_factory,
):
    """A test selection containing 2 geometries and one speed"""
    bc = {"Vx": 10}
    geometries = [
        geometry_factory(
            id="geom1",
            predictions=[prediction_factory(id="pred1", boundary_conditions=bc)],
        ),
        geometry_factory(
            id="geom2",
            predictions=[prediction_factory(id="pred2", boundary_conditions=bc)],
        ),
    ]
    yield Selection(geometries, [bc])


def test_selection_post_processing_global_coefficients(test_selection, httpx_mock):
    """WHEN I call post.post.global_coefficients() on a selection
    THEN the /GlobalCoefficients endpoint is called for each prediction in the selection
    AND I get a list of GlobalCoefficients objects in return
    """
    assert len(test_selection.points) == 2

    for num in [1, 2]:
        httpx_mock.add_response(
            method="POST",
            url=f"https://test.test/predictions/pred{num}/post-processings/GlobalCoefficients",
            json={"id": f"gc{num}", "state": "successful"},
            status_code=200,
        )

    post_processings = test_selection.post.global_coefficients()
    assert isinstance(post_processings, ExportablePPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, GlobalCoefficients)


def test_selection_post_processing_surface_evolution(test_selection, httpx_mock):
    """WHEN I call post.post.surface_evolution() on a selection
    THEN the /SurfaceEvol endpoint is called for each prediction in the selection
    AND I get a list of SurfaceEvolution objects in return
    """
    assert len(test_selection.points) == 2

    for num in [1, 2]:
        httpx_mock.add_response(
            method="POST",
            url=f"https://test.test/predictions/pred{num}/post-processings/SurfaceEvol",
            json={"id": f"se{num}", "state": "successful"},
            status_code=200,
        )
    post_processings = test_selection.post.surface_evolution(axis="x", delta=0.02)
    assert isinstance(post_processings, ExportablePPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, SurfaceEvolution)


def test_selection_post_processing_slice(test_selection, httpx_mock):
    """WHEN I call post.post.slice() on a selection
    THEN the /Slice endpoint is called for each prediction in the selection
    AND I get a list of Slice objects in return
    """
    assert len(test_selection.points) == 2

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred1/post-processings/Slice",
        json={"id": "slice01", "status": "queued"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred2/post-processings/Slice",
        json={"id": "slice02", "status": "queued"},
        status_code=200,
    )
    post_processings = test_selection.post.slice(axis="z", coordinate=30)
    assert isinstance(post_processings, PPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, Slice)


def test_selection_post_processing_volume_vtu(test_selection, httpx_mock):
    """WHEN I call post.post.volume_vtu() on a selection
    THEN the /VolumeVTU endpoint is called for each prediction in the selection
    AND I get a list of PostProcessingVTUExport objects in return
    """
    assert len(test_selection.points) == 2

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred1/post-processings/VolumeVTU",
        json={"id": "vtu01", "status": "queued"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred2/post-processings/VolumeVTU",
        json={"id": "vtu02", "status": "queued"},
        status_code=200,
    )
    post_processings = test_selection.post.volume_vtu()
    assert isinstance(post_processings, PPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, VolumeVTU)


def test_selection_post_processing_surface_vtp(test_selection, httpx_mock):
    """WHEN I call post.post.surface_vtp() on a selection
    THEN the /SurfaceVTP endpoint is called for each prediction in the selection
    AND I get a list of PostProcessingVTUExport objects in return
    """
    assert len(test_selection.points) == 2

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred1/post-processings/SurfaceVTP",
        json={"id": "se1"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred2/post-processings/SurfaceVTP",
        json={"id": "se3"},
        status_code=200,
    )
    post_processings = test_selection.post.surface_vtp()
    assert isinstance(post_processings, PPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, SurfaceVTP)


def test_selection_post_processing_surface_vtp_td_location(test_selection, httpx_mock):
    """WHEN I call post.post.surface_vtp() on a selection
    THEN the /SurfaceVTPTDLocation endpoint is called for each prediction in the selection
    AND I get a list of SurfaceVTPTDLocation objects in return
    """
    assert len(test_selection.points) == 2

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred1/post-processings/SurfaceVTPTDLocation",
        json={"id": "vtu01", "status": "queued"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred2/post-processings/SurfaceVTPTDLocation",
        json={"id": "vtu02", "status": "queued"},
        status_code=200,
    )
    post_processings = test_selection.post.surface_vtp_td_location()
    assert isinstance(post_processings, PPList)
    assert len(post_processings) == 2
    for pp in post_processings:
        assert isinstance(pp, SurfaceVTPTDLocation)


def test_selection_post_processing_error(test_selection, httpx_mock):
    """WHEN I call post.post.volume_vtu() on a selection
    AND the first VTU fails
    THEN the second VTU is still executed
    AND a ApiClientError is raised
    AND the working VTU has still been created
    """
    assert len(test_selection.points) == 2

    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred1/post-processings/VolumeVTU",
        json={
            "errors": {"json": {"Vx": "This is too fast"}},
            "status": "SppedingUp",
        },
        status_code=422,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/predictions/pred2/post-processings/VolumeVTU",
        json={"id": "vtu03", "status": "queued"},
        status_code=200,
    )
    with pytest.raises(ApiClientError):
        test_selection.post.volume_vtu()

    # Note: pytest-httpx doesn't provide an easy way to check call counts like responses
    # The test still verifies that the error is raised and the second VTU is created
    assert test_selection.predictions[1].post.volume_vtu().id == "vtu03"



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, NamedTuple, Optional, Union

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.model_configuration import ModelConfiguration
from ansys.simai.core.data.types import Identifiable, get_id_from_identifiable
from ansys.simai.core.errors import InvalidArguments, ProcessingError

if TYPE_CHECKING:
    from ansys.simai.core.data.global_coefficients_requests import (
        ProcessGlobalCoefficient,
    )
    from ansys.simai.core.data.models import Model
    from ansys.simai.core.data.training_data import TrainingData
    from ansys.simai.core.data.workspaces import Workspace

EXTRA_CALCULETTE_FIELDS = ["Area", "Normals", "Centroids"]

logger = logging.getLogger(__name__)


class IsTrainableInfo(NamedTuple):
    """Properties for project's trainability.

    The objects of this class can be used as booleans
    in condition statements as in the example:

    Example:
        Verify the project is trainable

        .. code-block:: python

            pt = my_project.is_trainable()

            if pt:
                print(pt)

        It prints:

        .. code-block:: shell

            <is_trainable: False, reason(s): Not enough data to train a model: we need at least 3 data points to train a model.>

    Attributes:
        is_trainable (bool):    True if the project is trainable, False if it is not.
        reason (str):           If not_trainable is False, the reason why the project is not trainable. None otherwise.

    """

    is_trainable: bool
    reason: str = None

    def __bool__(self) -> bool:
        return self.is_trainable

    def __repr__(self) -> str:
        return f"<is_trainable: {self.is_trainable}, reason(s): {self.reason}>"


@dataclass
class ContinuousLearningCapabilities:
    """Provides a project's continuous learning capabilities.

    Args:
        able: Is this project able to use continuous learning feature
        reasons: Reasons why a project can't use continuous learning feature
    """

    able: bool
    reasons: list[str]


@dataclass
class TrainingCapabilities:
    """Provides a project's training capabilities.

    Args:
        continuous_learning: Continuous learning capabilities.
    """

    continuous_learning: ContinuousLearningCapabilities


class Project(DataModel):
    """Provides the local representation of a project object."""

    def __repr__(self) -> str:
        return f"<Project: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of project."""
        return self.fields["name"]

    def rename(self, new_name: str) -> None:
        """Rename the project.

        Args:
            new_name: New name to give to the project.
        """
        self._client._api.update_project(self.id, name=new_name)
        self.reload()

    @property
    def data(self) -> list["TrainingData"]:
        """(**Deprecated**)List of all :class:`~ansys.simai.core.data.training_data.TrainingData` instances in the project.

        Use :py:meth:`.list_training_data` instead.
        """
        return self.list_training_data()

    @property
    def sample(self) -> Optional["TrainingData"]:
        """Sample of the project. The sample determines what variable and settings are available during model configuration."""
        raw_sample = self.fields["sample"]
        if raw_sample is None:
            return None
        return self._client.training_data._model_from(self.fields["sample"])

    @property
    def training_capabilities(self) -> TrainingCapabilities:
        """Training capabilities of the project. Determines whether a project can use continuous learning."""
        raw_tc = self.fields["training_capabilities"]
        continuous_learning = ContinuousLearningCapabilities(**raw_tc["continuous_learning"])
        return TrainingCapabilities(continuous_learning=continuous_learning)

    @sample.setter
    def sample(self, new_sample: Identifiable["TrainingData"]):
        td_id = get_id_from_identifiable(new_sample)
        self._client._api.set_project_sample(self.id, td_id)
        self.reload()

    @property
    def last_model_configuration(self) -> Optional[ModelConfiguration]:
        """Last :class:`configuration <ansys.simai.core.data.model_configuration.ModelConfiguration>` used for model training in this project."""
        raw_last_model_configuration = self.fields.get("last_model_configuration")
        if raw_last_model_configuration is None:
            return None
        return ModelConfiguration._from_payload(project=self, **raw_last_model_configuration)

    def list_workspaces(self) -> list["Workspace"]:
        """Lists all :class:`~ansys.simai.core.data.workspaces.Workspace` instances in the project."""
        workspaces = self._client._api.get_project_related_workspaces(self.id)
        return [self._client.workspaces._model_from(workspace) for workspace in workspaces]

    def list_training_data(self) -> list["TrainingData"]:
        """List of all :class:`~ansys.simai.core.data.training_data.TrainingData` instances in the project."""
        raw_td_list = self._client._api.iter_training_data_in_project(self.id)
        return [
            self._client.training_data._model_from(training_data) for training_data in raw_td_list
        ]

    def list_models(self) -> list["Model"]:
        """List of all :class:`~ansys.simai.core.data.models.Model` instances in the project."""
        models_data = self._client._api.get_project_models(self.id)
        return [self._client.models._model_from(model_data) for model_data in models_data]

    def is_trainable(self) -> IsTrainableInfo:
        """Check if the project meets the prerequisites to be trained."""
        tt = self._client._api.is_project_trainable(self.id)
        return IsTrainableInfo(**tt)

    def delete(self) -> None:
        """Delete the project."""
        self._client._api.delete_project(self.id)

    def get_variables(self) -> Optional[dict[str, list[str]]]:
        """Get the available variables for the model's input/output."""
        if not self.sample:
            return None

        sample_metadata = self.sample.fields.get("extracted_metadata")
        data = {}
        for key, vals in sample_metadata.items():
            local_fields = vals.get("fields", [])
            data[key] = [local_field.get("name") for local_field in local_fields]
        return data

    def process_gc_formula(
        self,
        gc_formula: str,
        bc: list[str] = None,
        surface_variables: list[str] = None,
        gc_location: Literal["cells", "points"] = "cells",
    ) -> Union[float, None]:
        """Process the formula of a global coefficient according to the project sample.
        It handles checking and computing the global coefficient formula as one workflow.
        """

        if not self.sample:
            raise ProcessingError(
                f"No sample is set for the project {self.id}. A sample should be set "
                f"for computing the results of a Global Coefficients formula."
            )

        sample_metadata = self.sample.fields.get("extracted_metadata")

        gc_process: ProcessGlobalCoefficient = (
            self._client._process_gc_formula_directory._model_from(
                data={
                    "id": f"{self.id}-process-{gc_formula}-{gc_location}",
                },
                project_id=self.id,
                gc_formula=gc_formula,
                sample_metadata=sample_metadata,
                bc=bc,
                surface_variables=surface_variables,
                gc_location=gc_location,
            )
        )

        gc_process.run()
        gc_process.wait()

        return gc_process.result

    def cancel_build(self):
        """Cancels a build if there is one pending."""

        self.reload()
        if self.fields.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_build(self.id)

    def set_as_current_project(self) -> None:
        """Configure the client to use this project instead of the one currently configured."""
        self._client.current_project = self


class ProjectDirectory(Directory[Project]):
    """Provides a collection of methods related to projects.

    This class is accessed through ``client.projects``.

    Example:
        List all projects::

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.projects.list()
    """

    _data_model = Project

    def list(self) -> list[Project]:
        """List all projects available on the server."""
        return [self._model_from(data) for data in self._client._api.projects()]

    def create(self, name: str) -> Project:
        """Create a project."""
        return self._model_from(self._client._api.create_project(name=name))

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> Project:
        """Get a project by either ID or name.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the project.
            name: Name of the project.

        Raises:
            ansys.simai.core.errors.NotFoundError: If the project doesn't exist
        """
        if name and id:
            raise InvalidArguments("Cannot specify both 'id' and 'name' arguments.")
        elif name:
            return self._model_from(self._client._api.get_project_by_name(name))
        elif id:
            return self._model_from(self._client._api.get_project(id))
        else:
            raise InvalidArguments("Either 'id' or 'name' argument must be specified.")

    def delete(self, project: Identifiable[Project]) -> None:
        """Delete a project.

        Args:
            project: ID or :class:`model <Project>` of the project.
        """
        self._client._api.delete_project(get_id_from_identifiable(project))

    def cancel_build(self, project: Identifiable[Project]):
        """Cancel a build if one is in progress.

        Args:
            project: ID or :class:`model <Project>` of the project.
        """

        project_id = get_id_from_identifiable(project)
        project_response = self._client._api.get_project(project_id)
        if project_response.get("is_being_trained") is False:
            raise ProcessingError("No build pending for this project.")
        self._client._api.cancel_build(project_id)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_basic_run_predictions:

Running Physics Predictions
================================================

This example demonstrates how to upload geometries to a workspace and run predictions
using a trained SimAI model. It also shows how to extract and save prediction results,
including global coefficients and confidence scores.

Before you begin
----------------------------------------------------

- Complete ":ref:`ref_basic_build_model`" to train a SimAI model.
- Ensure the model training completed successfully.
- Have a dataset folder with subdirectories containing geometry files. These geometry files can come from a Generative Design model.
- (Optional) Prepare boundary condition JSON files if your model requires them.

"""

###############################################################################
# Import necessary libraries
# ----------------------------------------------------

import json
import os

import ansys.simai.core as asc
from ansys.simai.core.data.geometries import Geometry
from ansys.simai.core.data.predictions import Prediction

###############################################################################
# Configure your settings
# ----------------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION_NAME = "<your_organization>"  # Replace with your organization name
WORKSPACE_NAME = "<your_workspace_name>"  # Replace with your workspace name
DATASET_PATH = "<your_dataset>"  # Path to your dataset directory

###############################################################################
# Initialize the client and set workspace
# ----------------------------------------------------
# Connect to SimAI and set the workspace to use for predictions:

simai_client = asc.SimAIClient(organization=ORGANIZATION_NAME)

###############################################################################
# Set the current workspace:

simai_client.set_current_workspace(WORKSPACE_NAME)
current_workspace = simai_client.current_workspace
print(f"Using workspace: {current_workspace.name}")

###############################################################################
# A workspace contains a trained model and is created when a model build completes successfully.

###############################################################################
# Prepare data structures
# ----------------------------------------------------
# Create lists to store geometries and their boundary conditions:

geometries: list[Geometry] = []
boundary_conditions: list[dict] = []

###############################################################################
# Upload geometries and load boundary conditions
# ----------------------------------------------------
# Process each subdirectory in the dataset:

for dir in os.listdir(f"{DATASET_PATH}"):
    print(f"Processing directory: {dir}")

    # Define the geometry file path
    # The file is uploaded with a specific name for easy identification
    file = (
        f"{DATASET_PATH}/{dir}/surface.vtp",  # Local file path
        f"{dir}.vtp",  # Name to use in SimAI
    )

    # Check if this geometry already exists in the workspace
    geoms = simai_client.geometries.list(
        workspace=current_workspace,
    )

    if f"{dir}.vtp" in [geom.name for geom in geoms]:
        print(f"Geometry {dir}.vtp already uploaded, skipping upload.")
        # Find and use the existing geometry
        existing_geom = [geom for geom in geoms if geom.name == f"{dir}.vtp"][0]
        geometries.append(existing_geom)
    else:
        print(f"Uploading geometry {dir}.vtp...")
        geom = simai_client.geometries.upload(file, workspace=current_workspace)
        geometries.append(geom)

    ###########################################################################
    # Load boundary conditions (if applicable)
    # ----------------------------------------------------
    # This section is optional and only needed if your model requires boundary conditions.
    # If your model does not use boundary conditions, you can skip this section.

    try:
        with open(f"{DATASET_PATH}/{dir}/boundary_conditions.json", "r") as f:
            print(f"Loading boundary conditions for {dir}...")
            boundary_conditions_for_geom = json.load(f)
            boundary_conditions.append(boundary_conditions_for_geom)
    except FileNotFoundError:
        print(f"No boundary conditions file found for {dir}, using empty boundary conditions.")
        boundary_conditions.append({})

###############################################################################
# Run predictions on all geometries
# ----------------------------------------------------
# Wait for all geometries to be processed, then run predictions:

preds: list[Prediction] = []

print("\nWaiting for geometries to be processed...")
for geom in geometries:
    print(f"Waiting for geometry {geom.name}...")
    geom.wait()
    print(f"Geometry {geom.name} is processed.")

    # Run prediction with boundary conditions if available
    if boundary_conditions[geometries.index(geom)]:
        preds.append(
            geom.run_prediction(boundary_conditions=boundary_conditions[geometries.index(geom)])
        )
        print(f"Prediction started for geometry {geom.name} with boundary conditions.")
    else:
        # Run prediction without boundary conditions
        preds.append(geom.run_prediction())
        print(f"Prediction started for geometry {geom.name}.")

###############################################################################
# Collect prediction results
# ----------------------------------------------------
# Wait for all predictions to complete and extract results:

results = {}

print("\nWaiting for predictions to complete...")
for pred in preds:
    geom_name = geometries[preds.index(pred)].name
    print(f"Waiting for prediction on geometry {geom_name}...")
    pred.wait()
    print(f"Prediction completed for geometry {geom_name}.")

    # Extract global coefficients and confidence score
    results[geom_name] = pred.post.global_coefficients().data
    results[geom_name]["confidence_score"] = pred.raw_confidence_score

###############################################################################
# Save results to file
# ----------------------------------------------------
# Export all prediction results to a JSON file for further analysis:

print("\nSaving results to results.json...")
with open("results.json", "w") as f:
    json.dump(results, f, indent=4)

print("Results saved successfully!")

###############################################################################
# Understanding the results
# ----------------------------------------------------
# The results dictionary contains:
#
# - Global coefficients: Scalar values computed from the prediction,
# - Confidence score: A measure of how familiar this prediction is for the model.
#
# You can also download full field results (VTP files) for visualization:
#
# - Use ``pred.post.surface_vtp()`` to get surface fields.
# - Use ``pred.post.volume_vtu()`` to get volume fields.

###############################################################################
# Example: Download surface VTP for visualization
# ----------------------------------------------------
# Run the following code to download the first prediction's surface VTP:

if preds:
    first_pred = preds[0]
    first_geom_name = geometries[0].name
    output_path = f"{first_geom_name}_result.vtp"
    first_pred.post.surface_vtp().data.download(output_path)
    print(f"Downloaded surface VTP to {output_path}")

###############################################################################
# Next steps
# ----------------------------------------------------
# With your prediction results, you can:
#
# - Analyze global coefficients to evaluate design performance.
# - Visualize field results in ParaView or similar tools.
# - Compare predictions against validation data.
# - Use results to guide design optimization.



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.simai.core.errors import InvalidArguments


def test_get_geomai_training_data_by_id(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geomai/training-data/td-123",
        json={"id": "td-123", "name": "My Training Data"},
        status_code=200,
    )
    td = simai_client.geomai.training_data.get(id="td-123")
    assert td.id == "td-123"
    assert td.name == "My Training Data"


def test_get_geomai_training_data_by_name(simai_client, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geomai/training-data/name/My%20Training%20Data",
        json={"id": "td-456", "name": "My Training Data"},
        status_code=200,
    )
    td = simai_client.geomai.training_data.get(name="My Training Data")
    assert td.id == "td-456"
    assert td.name == "My Training Data"


def test_get_geomai_training_data_invalid_arguments(simai_client):
    with pytest.raises(InvalidArguments) as e:
        simai_client.training_data.get()
    assert str(e.value) == "Either 'id' or 'name' argument must be specified."

    with pytest.raises(InvalidArguments) as e:
        simai_client.training_data.get(id="td-123", name="My Training Data")
    assert str(e.value) == "Cannot specify both 'id' and 'name' arguments."



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_basic_build_model:

Building a SimAI Model
============================================

This example demonstrates how to configure a SimAI model with inputs, outputs, global coefficients,
and domain of analysis, then start the model training process.

Before you begin
-------------------------------------------

- Complete ":ref:`ref_basic_create_project_upload_data`" to create a project with training data.
- Ensure all training data in your project are ready (processed successfully).
- Know the names of surfaces and boundary conditions in your training data.
- Determine which global coefficients you want to calculate.

"""

###############################################################################
# Import necessary libraries
# -------------------------------------------

import ansys.simai.core as asc
from ansys.simai.core.data.model_configuration import (
    DomainOfAnalysis,
    GlobalCoefficientDefinition,
    ModelConfiguration,
    ModelInput,
    ModelOutput,
)

###############################################################################
# Configure your settings
# -------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION_NAME = "<your_organization>"  # Replace with your organization name
PROJECT_NAME = "<your_project_name>"  # Replace with your project name

###############################################################################
# Initialize the client and get the project
# -------------------------------------------
# Connect to SimAI:

simai_client = asc.SimAIClient(organization=ORGANIZATION_NAME)

###############################################################################
# Retrieve the project by name:

project = simai_client.projects.get(name=PROJECT_NAME)
print(f"Using existing project: {PROJECT_NAME}")

###############################################################################
# Configure model inputs
# -------------------------------------------
# Define which surfaces and boundary conditions the model will use as inputs.
# Replace the placeholder names with the actual names from your training data:

model_input = ModelInput(
    surface=["<your_surface_name>"],  # List of surface names to use as input
    boundary_conditions=["<your_BC_name>"],  # List of boundary condition names
)

###############################################################################
# Configure model outputs
# -------------------------------------------
# Define which surfaces the model will predict.
# These are typically the same surfaces as the inputs, but you can specify different ones:

model_output = ModelOutput(
    surface=["<your_surface_name>"],  # List of surface names to predict
)

###############################################################################
# Define global coefficients
# -------------------------------------------
#
# Global coefficients are scalar values calculated from the prediction results.
# They can be used to extract key performance indicators from your simulations.
#
# In this example, we calculate the integral of a field called "Photometric".
# You can define multiple global coefficients with different formulas.
#
# Available locations: ``"points"`` or ``"cells"``.

global_coefficients = [
    GlobalCoefficientDefinition(
        name="<your_global_coefficient_name>",  # Name for this global coefficient
        formula="integral(Photometric)",  # Formula to calculate the global coefficient value
        gc_location="points",  # Calculate at points (or at "cells")
    )
]

###############################################################################
# Define domain of analysis
# -------------------------------------------
# The domain of analysis specifies the spatial extent that the model will analyze.
# Values can be specified as absolute dimensions or relative to minimum values.
#
# Format: reference_type, min_value, max_value
#
# - ``reference_type``: ``"relative_to_min"``, ``"relative_to_max"``, ``"relative_to_center``" or ``"absolute"``.
# - ``min_value``: minimum dimension value.
# - ``max_value``: maximum dimension value.
#
# You can find appropriate values by analyzing your training data in the SimAI platform.

doa = DomainOfAnalysis(
    length=("relative_to_min", 0.102, 1.219),  # Length bounds
    width=("relative_to_min", 0.102, 1.219),  # Width bounds
    height=("relative_to_min", 0.134, 1.61),  # Height bounds
)

###############################################################################
# Create model configuration
# -------------------------------------------
# Combine all the configuration elements into a ModelConfiguration object.
#
# Build presets determine the training duration:
#
# - ``debug``: Quick training for testing.
# - ``1_day``: Short training.
# - ``2_days``: Standard training (recommended).
# - ``7_days``: Maximum training for best accuracy.


mdl_conf = ModelConfiguration(
    project=project,  # Project containing the training data
    build_preset="1_day",  # Duration of the build
    build_on_top=False,  # Start training from scratch
    input=model_input,  # Model input configuration
    output=model_output,  # Model output configuration
    global_coefficients=global_coefficients,  # Global coefficient definitions
    domain_of_analysis=doa,  # Domain of analysis bounds
)

###############################################################################
# Verify and build the model
# -------------------------------------------
# Before building, check if the project meets all requirements for training:

if project.is_trainable():
    print("Project is trainable. Starting model build...")
    new_model = simai_client.models.build(mdl_conf)
    print(f"Model build started with ID: {new_model.id}")
    print(
        "The model will take some time to train. You can monitor its progress in the SimAI platform."
    )
else:
    print("Project is not trainable. Please check the following:")
    print("- All training data must be processed successfully.")
    print("- The project must have enough training data.")
    print("- All required surfaces and boundary conditions must exist in the training data.")

###############################################################################
# Next steps
# -------------------------------------------
# Once your model is trained, you can:
#
# - Monitor the training progress in the SimAI platform.
# - Run predictions on new geometries: :ref:`ref_basic_run_predictions`.
# - Evaluate model performance using validation data.



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import platform
import time
from pathlib import Path

import pytest

from ansys.simai.core.utils.files import get_cache_dir


@pytest.mark.skipif(
    platform.system() != "Linux",
    reason="This test uses XDG_CACHE_HOME to avoid touching the host system",
)
def test_get_cache_dir_linux(tmpdir, monkeypatch):
    monkeypatch.setenv("XDG_CACHE_HOME", str(tmpdir))
    cache_dir = get_cache_dir()
    expected_cache_dir = Path(tmpdir).absolute() / "ansys" / "pysimai"
    assert cache_dir.absolute() == expected_cache_dir
    old_file = cache_dir / "old"
    new_file = cache_dir / "new"
    new_file.touch()
    old_file.touch()
    now = time.time()
    a_year_ago = now - 30000000
    os.utime(old_file, times=(now, a_year_ago))
    cache_dir = get_cache_dir()
    assert new_file.is_file()
    assert not old_file.is_file()



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.simai.core.data.geometries import Geometry


def test_prediction_failure(simai_client):
    """WHEN creating a Prediction in an error state and with an "error" field
    THEN prediction.has_failed is true and failure_reason describes the error
    """
    prediction = simai_client._prediction_directory._model_from(
        {"id": "pizza-4789", "state": "rejected", "error": "The pizza was lukewarm"}
    )
    assert prediction.has_failed
    assert prediction.failure_reason == "The pizza was lukewarm"

    prediction = simai_client._prediction_directory._model_from(
        {"id": "drink-4673", "state": "failure", "error": "Drink was spilled over"}
    )
    assert prediction.has_failed
    assert prediction.failure_reason == "Drink was spilled over"


def test_prediction_delete(simai_client, httpx_mock):
    """WHEN deleting a Prediction
    THEN a DELETE query on predictions/id endpoint is called
    ALSO the prediction doesn't exist anymore in its Directory's registry
    """
    pred_id = "uninteresting-prediction-001"
    prediction = simai_client._prediction_directory._model_from(
        {"id": pred_id, "state": "successful"}
    )

    assert simai_client._prediction_directory._registry[pred_id] == prediction

    httpx_mock.add_response(
        method="DELETE",
        url=f"https://test.test/predictions/{pred_id}?confirm=true",
        status_code=204,
    )
    prediction.delete()

    assert pred_id not in simai_client._prediction_directory._registry


def test_prediction_wait_deleted(simai_client, httpx_mock):
    """WHEN deleting a Prediction that is not done
    THEN prediction.wait() immediately returns
    """
    pred_id = "uninteresting-prediction-002"
    prediction = simai_client._prediction_directory._model_from({"id": pred_id, "state": "pending"})

    httpx_mock.add_response(
        method="DELETE",
        url=f"https://test.test/predictions/{pred_id}?confirm=true",
        status_code=204,
    )
    prediction.delete()
    prediction.wait()


def test_prediction_geometry_attribute(prediction_factory, httpx_mock):
    """WHEN accessing a Prediction's geometry attribute
    THEN the geometry object is returned
    """
    prediction = prediction_factory(geometry_id="mexico")
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/mexico",
        json={"id": "mexico", "state": "successful"},
        status_code=200,
    )

    assert isinstance(prediction.geometry, Geometry)
    assert prediction.geometry.id == "mexico"


def test_prediction_call_geometry_attribute_twice(prediction_factory, httpx_mock):
    """WHEN accessing twice the geometry attribute of a Prediction
    THEN the endpoint is called only once
    """
    prediction = prediction_factory(geometry_id="cuba")
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/cuba",
        json={"id": "cuba", "state": "successful"},
        status_code=200,
    )

    prediction.geometry  # noqa: B018
    prediction.geometry  # noqa: B018

    assert len(httpx_mock.get_requests()) == 1


def test_prediction_call_geometry_attribute_already_registered(
    geometry_factory, prediction_factory, httpx_mock
):
    """WHEN accessing the geometry attribute of a prediction when the geometry exists locally
    THEN no query is ran
    """
    geometry_factory(id="registered_geometry")  # Registering geometry
    prediction = prediction_factory(geometry_id="registered_geometry")

    prediction.geometry  # noqa: B018
    assert len(httpx_mock.get_requests()) == 0


def test_run(simai_client, geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/geom-0",
        json={"id": "geom-0"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    simai_client.predictions.run(geometry.id, Vx=10.5)


def test_run_dict_bc(simai_client, geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/geom-0",
        json={"id": "geom-0"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    simai_client.predictions.run(geometry.id, {"Vx": 10.5})


def test_run_no_bc(simai_client, geometry_factory, httpx_mock):
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/geometries/geom-0",
        json={"id": "geom-0"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="POST",
        url="https://test.test/geometries/geom-0/predictions",
        json={"id": "pred-0"},
        status_code=200,
    )
    geometry = geometry_factory(id="geom-0")
    simai_client.predictions.run(geometry.id)


def test_confidence_score(prediction_factory, httpx_mock):
    """WHEN accessing a Prediction's confidence score properties
    THEN the corresponding values are returned
    """
    prediction = prediction_factory(confidence_score="high", raw_confidence_score=0.94107)
    empty_prediction = prediction_factory()
    bad_prediction = prediction_factory(confidence_score="abysmal")

    assert prediction.confidence_score == "high"
    assert prediction.raw_confidence_score == 0.94107
    assert empty_prediction.confidence_score is None
    assert empty_prediction.raw_confidence_score is None
    with pytest.raises(ValueError) as exc:
        assert bad_prediction.confidence_score
    assert str(exc.value) == "Must be None or one of: 'high', 'low', None."



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import logging
from typing import TYPE_CHECKING

from ansys.simai.core.data.geomai.models import GeomAIModelDirectory
from ansys.simai.core.data.geomai.predictions import GeomAIPredictionDirectory
from ansys.simai.core.data.geomai.projects import GeomAIProject, GeomAIProjectDirectory
from ansys.simai.core.data.geomai.training_data import GeomAITrainingDataDirectory
from ansys.simai.core.data.geomai.training_data_parts import GeomAITrainingDataPartDirectory
from ansys.simai.core.data.geomai.workspaces import GeomAIWorkspace, GeomAIWorkspaceDirectory
from ansys.simai.core.errors import (
    InvalidClientStateError,
    InvalidConfigurationError,
    NotFoundError,
)
from ansys.simai.core.utils.configuration import ClientConfig

if TYPE_CHECKING:
    from ansys.simai.core.client import SimAIClient


logger = logging.getLogger(__name__)


class GeomAIClient:
    def __init__(self, simai_client: "SimAIClient", config: ClientConfig):
        self.client = simai_client
        self._model_directory = GeomAIModelDirectory(client=simai_client)
        self._prediction_directory = GeomAIPredictionDirectory(client=simai_client)
        self._project_directory = GeomAIProjectDirectory(client=simai_client)
        self._workspace_directory = GeomAIWorkspaceDirectory(client=simai_client)
        self._training_data_directory = GeomAITrainingDataDirectory(client=simai_client)
        self._training_data_part_directory = GeomAITrainingDataPartDirectory(client=simai_client)
        self._current_workspace = None
        self._current_project = None
        if config.geomai_workspace is not None:
            self.set_current_workspace(config.geomai_workspace)
        if config.geomai_project is not None:
            self.set_current_project(config.geomai_project)

    @property
    def current_workspace(self) -> GeomAIWorkspace:
        """Workspace currently used by the GeomAI client.

        Note:
            You should not set the workspace directly. Instead, use the :meth:`set_current_workspace`
            method, which uses the workspace name and ensures that the workspace exists.
        """
        if self._current_workspace is None:
            raise InvalidClientStateError("Current workspace is not set.")
        return self._current_workspace

    @current_workspace.setter
    def current_workspace(self, workspace: GeomAIWorkspace):
        self._current_workspace = workspace

    def set_current_workspace(self, name: str):
        """Set the current workspace for the GeomAI client.

        Args:
            name: Name of the workspace that the client should switch to.
        """
        try:
            # Ensure the workspace exists
            workspace = self.workspaces.get(name=name)
            logger.info(f"Workspace set to {name}.")
        except NotFoundError:
            raise InvalidConfigurationError(
                f"""Configured workspace {name} does not exist on the server.
{self._available_workspaces_string}"""
            ) from None
        self._current_workspace = workspace

    @property
    def _available_workspaces_string(self):
        available_workspaces = [workspace.name for workspace in self.workspaces.list()]
        return f"Available workspaces are: {available_workspaces}."

    @property
    def current_project(self) -> GeomAIProject:
        """Project currently used by the GeomAI client.

        Note:
            You should not set the project directly. Instead, use the :meth:`set_current_project`
            method, which uses the project name and ensures that the project exists.
        """
        if self._current_project is None:
            raise InvalidClientStateError("Current project is not set.")
        return self._current_project

    @current_project.setter
    def current_project(self, project: GeomAIProject):
        self._current_project = project

    def set_current_project(self, name: str):
        """Set the current project for the GeomAI client.

        This method affects how some methods related to projects or associated
        data behave.

        Args:
            name: Name of the project that the client should switch to.
        """
        try:
            # Ensure the project exists
            project = self.projects.get(name=name)
            logger.info(f"Project set to {name}")
        except NotFoundError:
            raise InvalidConfigurationError(
                f"""Configured project {name} does not exist on the server.
{self._available_projects_string}"""
            ) from None
        self._current_project = project

    @property
    def _available_projects_string(self):
        available_projects = [project.name for project in self.projects.list()]
        return f"Available projects are: {available_projects}."

    @property
    def training_data(self) -> GeomAITrainingDataDirectory:
        """Representation of all GeomAI training data on the server.
        For more information, see :ref:`geomai_training_data`.
        """
        return self._training_data_directory

    @property
    def training_data_parts(self) -> GeomAITrainingDataPartDirectory:
        """Representation of all GeomAI training data parts on the server.
        For more information, see :ref:`geomai_training_data_parts`.
        """
        return self._training_data_part_directory

    @property
    def projects(self) -> GeomAIProjectDirectory:
        """Representation of all GeomAI projects on the server.
        For more information, see :ref:`geomai_projects`.
        """
        return self._project_directory

    @property
    def models(self) -> GeomAIModelDirectory:
        """Representation of all GeomAI models on the server.
        For more information, see :ref:`geomai_models`.
        """
        return self._model_directory

    @property
    def workspaces(self) -> GeomAIWorkspaceDirectory:
        """Representation of all GeomAI workspaces on the server.
        For more information, see :ref:`geomai_workspaces`.
        """
        return self._workspace_directory

    @property
    def predictions(self) -> GeomAIPredictionDirectory:
        """Representation of all GeomAI predictions on the server.
        For more information, see :ref:`geomai_predictions`.
        """
        return self._prediction_directory



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING, Dict, List, Optional

from ansys.simai.core.data.base import ComputableDataModel, Directory
from ansys.simai.core.data.types import (
    Filters,
    Identifiable,
    MonitorCallback,
    NamedFile,
    SizedIterator,
    get_id_from_identifiable,
    to_raw_filters,
    unpack_named_file,
)
from ansys.simai.core.errors import InvalidArguments
from ansys.simai.core.utils.pagination import DataModelIterator

if TYPE_CHECKING:
    from ansys.simai.core.data.geomai.projects import GeomAIProject
    from ansys.simai.core.data.geomai.training_data_parts import GeomAITrainingDataPart


def _upload_geomai_training_data_part(id, named_part, client, monitor_callback):
    with unpack_named_file(named_part) as (file, name, extension):
        (training_data_part_fields, upload_id) = client._api.create_geomai_training_data_part(
            id, name, extension
        )
        training_data_part = client.geomai.training_data_parts._model_from(
            training_data_part_fields, is_upload_complete=False
        )
        parts = client._api.upload_parts(
            f"geomai/training-data-parts/{training_data_part.id}/part",
            file,
            upload_id,
            monitor_callback=monitor_callback,
        )
        client._api.complete_geomai_training_data_part_upload(
            training_data_part.id, upload_id, parts
        )
    return training_data_part


class GeomAITrainingData(ComputableDataModel):
    """Provides the local representation of a training data object."""

    def __repr__(self) -> str:
        return f"<GeomAITrainingData: {self.id}, {self.name}>"

    @property
    def name(self) -> str:
        """Name of the training data."""
        return self.fields["name"]

    @property
    def parts(self) -> List["GeomAITrainingDataPart"]:
        """List of all :class:`parts<ansys.simai.core.data.geomai.training_data_parts.GeomAITrainingDataPart>`
        objects in the training data.
        """
        return [
            self._client.geomai.training_data_parts._model_from(training_data_part)
            for training_data_part in self.fields["parts"]
        ]

    @property
    def extracted_metadata(self) -> Optional[Dict]:
        """Metadata extracted from the training data."""
        return self.fields["extracted_metadata"]

    def extract_data(self) -> None:
        """Extract or reextract the data from a training data.

        Data should be extracted from a training data once all its parts have been fully uploaded.
        This is done automatically when using :meth:`GeomAITrainingDataDirectory.create_from_file` to create training data.

        Data can only be reextracted from a training data if the extraction previously failed or if new files have been added.
        """
        self._client._api.compute_geomai_training_data(self.id)

    def delete(self) -> None:
        """Delete the training data on the server."""
        self._client._api.delete_geomai_training_data(self.id)

    def upload_part(
        self, file: NamedFile, monitor_callback: Optional[MonitorCallback] = None
    ) -> "GeomAITrainingDataPart":
        """Add a part to the training data.

        Args:
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            monitor_callback: Optional callback for monitoring the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            Created :class:`~ansys.simai.core.data.geomai.training_data_parts.GeomAITrainingDataPart`.
        """
        return _upload_geomai_training_data_part(self.id, file, self._client, monitor_callback)

    def add_to_project(self, project: Identifiable["GeomAIProject"]):
        """Add the training data to a :class:`~.projects.GeomAIProject` object.

        Args:
            project: ID or :class:`model <.projects.GeomAIProject>` object of the project to add the data to.
        """
        project_id = get_id_from_identifiable(project)
        self._client._api.add_geomai_training_data_to_project(self.id, project_id)

    def remove_from_project(self, project: Identifiable["GeomAIProject"]):
        """Remove the training data from a :class:`~.projects.GeomAIProject` object.

        Args:
            project: ID or :class:`model <.projects.GeomAIProject>` of the project to remove data from.

        Raises:
            ansys.simai.core.errors.ApiClientError: If the data is the project's sample.
            ansys.simai.core.errors.ApiClientError: If the project is in training.
        """
        project_id = get_id_from_identifiable(project)
        self._client._api.remove_geomai_training_data_from_project(self.id, project_id)


class GeomAITrainingDataDirectory(Directory[GeomAITrainingData]):
    """Provides a collection of methods related to training data.

    This class is accessed through ``client.training_data``.

    Example:
        List all of the GeomAI training data::

            import ansys.simai.core as asc

            simai_client = asc.from_config()
            simai_client.geomai.training_data.list()
    """

    _data_model = GeomAITrainingData

    def iter(self, filters: Optional[Filters] = None) -> SizedIterator[GeomAITrainingData]:
        """Iterate over all :class:`GeomAITrainingData` objects on the server.

        Args:
            filters: Optional :obj:`~ansys.simai.core.data.types.Filters` to apply.

        Returns:
            Iterator over all :class:`GeomAITrainingData` objects on the server.
        """
        raw_filters = to_raw_filters(filters)
        raw_iterable = self._client._api.iter_geomai_training_data(filters=raw_filters)
        return DataModelIterator(raw_iterable, self)

    def list(self, filters: Optional[Filters] = None) -> List[GeomAITrainingData]:
        """List all :class:`GeomAITrainingData` objects on the server.

        Warning:
            This can take a very long time, consider using :py:meth:`~iter` instead.

        Args:
            filters: Optional :obj:`~ansys.simai.core.data.types.Filters` to apply.

        Returns:
            List of all :class:`GeomAITrainingData` objects on the server.
        """
        return list(self.iter(filters))

    def get(self, id: Optional[str] = None, name: Optional[str] = None) -> GeomAITrainingData:
        """Get a specific :class:`GeomAITrainingData` object from the server.

        You can specify either the ID or the name, not both.

        Args:
            id: ID of the training data.
            name: Name of the training data.

        Returns:
            :class:`GeomAITrainingData`

        Raises:
            NotFoundError: No training data with the given ID exists.
        """
        if name and id:
            raise InvalidArguments("Cannot specify both 'id' and 'name' arguments.")
        elif name:
            return self._model_from(self._client._api.get_geomai_training_data_by_name(name))
        elif id:
            return self._model_from(self._client._api.get_geomai_training_data(id))
        else:
            raise InvalidArguments("Either 'id' or 'name' argument must be specified.")

    def delete(self, training_data: Identifiable[GeomAITrainingData]) -> None:
        """Delete a :class:`GeomAITrainingData` object and its associated parts from the server.

        Args:
            training_data: ID or :class:`model <GeomAITrainingData>` object of the :class:`GeomAITrainingData` object.
        """
        return self._client._api.delete_geomai_training_data(
            get_id_from_identifiable(training_data)
        )

    def create(
        self, name: str, project: Optional[Identifiable["GeomAIProject"]] = None
    ) -> GeomAITrainingData:
        """Create an empty :class:`GeomAITrainingData` object.

        Args:
            name: Name to give the new :class:`GeomAITrainingData` object.
            project: :class:`~.projects.Project` object to associate the data with.

        Returns:
            Created :class:`GeomAITrainingData` object.

        See Also:
            :meth:`create_from_file` handles all the creation of training data and parts for you
        """
        project_id = get_id_from_identifiable(project, required=False)
        return self._model_from(self._client._api.create_geomai_training_data(name, project_id))

    def create_from_file(
        self,
        file: NamedFile,
        project: Optional[Identifiable["GeomAIProject"]] = None,
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> GeomAITrainingData:
        """Convenience function to create a :class:`GeomAITrainingData` object from a file.

        Args:
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            project: Optional project to attach the training data to.
                Defaults to the configured project if any.
            monitor_callback: Optional callback for monitoring the progress of the download.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.
        """
        project_id = get_id_from_identifiable(
            project, default=self._client.geomai._current_project, required=False
        )
        with unpack_named_file(file) as (file_bytes, name, extension):
            training_data = self._model_from(
                self._client._api.create_geomai_training_data(name, project_id)
            )
            _upload_geomai_training_data_part(
                training_data.id, file, self._client, monitor_callback
            )
            self._client._api.compute_geomai_training_data(training_data.id)
            return training_data

    def upload_part(
        self,
        training_data: Identifiable[GeomAITrainingData],
        file: NamedFile,
        monitor_callback: Optional[MonitorCallback] = None,
    ) -> "GeomAITrainingDataPart":
        """Add a part to a :class:`GeomAITrainingData` object.

        Args:
            training_data: ID or :class:`model <GeomAITrainingData>` object of the training data to
                add the part to.
            file: :obj:`~ansys.simai.core.data.types.NamedFile` to upload.
            monitor_callback: Optional callback for monitoring the progress of the upload.
                For more information, see the :obj:`~ansys.simai.core.data.types.MonitorCallback`
                object.

        Returns:
            Added :class:`~ansys.simai.core.data.geomai.training_data_parts.GeomAITrainingDataPart` object.
        """
        return _upload_geomai_training_data_part(
            get_id_from_identifiable(training_data),
            file,
            self._client,
            monitor_callback,
        )



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import math

import httpx
import pytest

from ansys.simai.core.data.post_processings import (
    GlobalCoefficients,
    Slice,
    SurfaceEvolution,
    VolumeVTU,
)
from ansys.simai.core.errors import ApiClientError


def test_post_processing_async_status(prediction_factory, httpx_mock):
    """WHEN Running a post-processing on a prediction
    THEN a PostProcessing object is returned, in status loading and not failed
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json={"id": "0123456", "state": "queued"},
        status_code=200,
    )

    global_coefficients = pred.post.global_coefficients()
    assert isinstance(global_coefficients, GlobalCoefficients)
    assert global_coefficients.id == "0123456"
    assert not global_coefficients.has_failed
    assert global_coefficients.failure_reason is None
    assert global_coefficients.is_pending


def test_post_processing_global_coefficients(prediction_factory, httpx_mock):
    """WHEN Running a GlobalCoefficients post-processing on a prediction
    THEN a POST request is made on the post-processings/GlobalCoefficients endpoint
    ALSO subsequent calls do not generate calls to the endpoint
    """

    pred = prediction_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/GlobalCoefficients",
        json={"id": "7777"},
        status_code=200,
    )

    global_coefficients = pred.post.global_coefficients()
    assert isinstance(global_coefficients, GlobalCoefficients)
    assert global_coefficients.id == "7777"

    pred.post.global_coefficients()
    assert len(httpx_mock.get_requests()) == 1


def test_post_processing_vtu(prediction_factory, httpx_mock):
    """WHEN Running a VTU  post-processing on a prediction
    THEN a POST request is made on the post-processings/VolumeVTU endpoint
    ALSO subsequent calls do not generate calls to the endpoint
    """
    pred = prediction_factory()
    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/VolumeVTU",
        json={"id": "6666"},
        status_code=200,
    )

    volume_vtu = pred.post.volume_vtu()
    assert isinstance(volume_vtu, VolumeVTU)
    assert volume_vtu.id == "6666"

    pred.post.volume_vtu()
    assert len(httpx_mock.get_requests()) == 1


def test_post_processing_surface_evolution(prediction_factory, httpx_mock):
    """WHEN Running a SurfaceEvolution post-processing on a prediction
    THEN a POST request is made on the post-processings/SurfaceEvol endpoint
    AND on subsequent access, the endpoint is not called
    """

    pred = prediction_factory()

    def request_callback(request: httpx.Request):
        payload = json.loads(request.content)
        if payload["axis"] == "x" and payload["delta"] == 5:
            return httpx.Response(200, json={"id": "7894"})
        if payload["axis"] == "y" and payload["delta"] == 9.5:
            return httpx.Response(200, json={"id": "1111"})
        raise Exception("This request is not expected in this test")

    for _ in range(2):
        httpx_mock.add_callback(
            request_callback,
            method="POST",
            url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol",
        )

    surface_evolution = pred.post.surface_evolution(axis="x", delta=5)
    assert isinstance(surface_evolution, SurfaceEvolution)
    assert surface_evolution.id == "7894"

    surface_evolution = pred.post.surface_evolution(axis="y", delta=9.5)
    assert isinstance(surface_evolution, SurfaceEvolution)
    assert surface_evolution.id == "1111"

    # call the post-processings with same parameters again
    surface_evolution = pred.post.surface_evolution(axis="x", delta=5)
    assert surface_evolution.id == "7894"

    surface_evolution = pred.post.surface_evolution(axis="y", delta=9.5)
    assert surface_evolution.id == "1111"

    # Check the URL has been called exactly twice, only once by parameter set
    assert len(httpx_mock.get_requests()) == 2


def test_post_processing_surface_evolution_parameters_values(prediction_factory, httpx_mock):
    """WHEN Running a SurfaceEvolution post-processing with good parameters,
    THEN a POST request is made on the post-processings/SurfaceEvol endpoint
    AND a PostProcessing object is returned
    """
    pred = prediction_factory()
    for _ in range(6):
        httpx_mock.add_response(
            method="POST",
            url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol",
            json={"id": "4444"},
            status_code=200,
        )

    surface_evolution = pred.post.surface_evolution(axis="x", delta=5)
    assert isinstance(surface_evolution, SurfaceEvolution)
    surface_evolution = pred.post.surface_evolution(axis="y", delta=5)
    assert isinstance(surface_evolution, SurfaceEvolution)
    surface_evolution = pred.post.surface_evolution(axis="z", delta=5)
    assert isinstance(surface_evolution, SurfaceEvolution)
    surface_evolution = pred.post.surface_evolution(axis="x", delta=1)
    assert isinstance(surface_evolution, SurfaceEvolution)
    surface_evolution = pred.post.surface_evolution(axis="x", delta=100)
    assert isinstance(surface_evolution, SurfaceEvolution)
    surface_evolution = pred.post.surface_evolution(axis="x", delta=0.0001)
    assert isinstance(surface_evolution, SurfaceEvolution)


def test_post_processing_surface_evolution_with_wrong_parameters(prediction_factory, httpx_mock):
    """WHEN Running a SurfaceEvolution PP with missing or wrong parameters,
    THEN a TypeError exception is raised
    """
    pred = prediction_factory()
    with pytest.raises(TypeError):
        pred.post.surface_evolution()
    with pytest.raises(TypeError):
        pred.post.surface_evolution(axis="x")
    with pytest.raises(TypeError):
        pred.post.surface_evolution(delta=5)
    with pytest.raises(TypeError):
        pred.post.surface_evolution(axis="y", delta=-10)
    with pytest.raises(TypeError):
        pred.post.surface_evolution(axis="y", delta="10")
    with pytest.raises(TypeError):
        pred.post.surface_evolution(axis="b", delta=10)
    with pytest.raises(TypeError):
        pred.post.surface_evolution(plane=[1, 2, 3, 4])


def test_post_processing_slice(prediction_factory, httpx_mock):
    """WHEN Running a Slice post-processing on a prediction
    THEN a POST request is made on the post-processings/Slice endpoint
    AND on subsequent access, the endpoint is not called
    """
    pred = prediction_factory()

    def request_callback(request):
        payload = json.loads(request.content)
        plane = payload["plane"]
        if plane == [1, 0, 0, 30.5]:
            return httpx.Response(200, json={"id": "976544"})
        if plane == [0, 1, 0, 4]:
            return httpx.Response(200, json={"id": "114455"})
        raise Exception("unexpected plane")

    for _ in range(2):
        httpx_mock.add_callback(
            request_callback,
            method="POST",
            url=f"https://test.test/predictions/{pred.id}/post-processings/Slice",
        )

    slice = pred.post.slice(axis="x", coordinate=30.5)
    assert isinstance(slice, Slice)
    assert slice.id == "976544"

    slice = pred.post.slice(axis="y", coordinate=4)
    assert isinstance(slice, Slice)
    assert slice.id == "114455"

    # Check the URL has been called exactly twice, only once by parameter set
    assert len(httpx_mock.get_requests()) == 2


def test_post_processing_slice_parameters_values(prediction_factory, httpx_mock):
    """WHEN Running a Slice post-processing with good parameters,
    THEN a POST request is made on the post-processings/Slice endpoint
    AND a PostProcessing object is returned
    """
    pred = prediction_factory()

    for _ in range(2):
        httpx_mock.add_response(
            method="POST",
            url=f"https://test.test/predictions/{pred.id}/post-processings/Slice",
            json={"id": "7654198"},
            status_code=200,
        )
    slice = pred.post.slice(axis="x", coordinate=4)
    assert isinstance(slice, Slice)
    slice = pred.post.slice(axis="z", coordinate=math.pi)
    assert isinstance(slice, Slice)


def test_post_processing_slice_with_wrong_parameters(prediction_factory, httpx_mock):
    """WHEN Running a Slice PP with missing or wrong parameters,
    THEN a ValueError or TypeError exception is raised
    """
    pred = prediction_factory()
    with pytest.raises(ValueError):
        pred.post.slice(axis="a", coordinate=20.4)
    with pytest.raises(TypeError):
        pred.post.slice(axis="x")
    with pytest.raises(TypeError):
        pred.post.slice(coordinate=20.4)


def test_post_processing_request_failure_raises_exception(prediction_factory, httpx_mock):
    """WHEN Running a post-processing
    IF back-end replies with an error status code
    THEN a ApiClientError is raised
    """
    pred = prediction_factory()

    httpx_mock.add_response(
        method="POST",
        url=f"https://test.test/predictions/{pred.id}/post-processings/SurfaceEvol",
        json={"error": "Something went wrong"},
        status_code=400,
    )
    with pytest.raises(ApiClientError):
        pred.post.surface_evolution(axis="x", delta=45)


def test_post_processing_reload(simai_client, httpx_mock):
    """WHEN Reloading a post-processing
    THEN a query is made to the post processing endpoint
    """
    pp_json = {"id": "88x88x", "type": "GlobalCoefficients"}
    pp = simai_client._post_processing_directory._model_from(data=pp_json)

    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/88x88x",
        json=pp_json,
        status_code=200,
    )
    pp.reload()


def test_post_processing_get(simai_client, httpx_mock):
    """WHEN Calling get() on post-processing directory
    THEN a post-processing object of corresponding type is returned
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/0001",
        json={"id": "0001", "type": "GlobalCoefficients"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/0002",
        json={"id": "0002", "type": "SurfaceEvolution"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/0003",
        json={"id": "0003", "type": "Slice"},
        status_code=200,
    )
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/0004",
        json={"id": "0004", "type": "VolumeVTU"},
        status_code=200,
    )
    assert isinstance(
        simai_client._post_processing_directory.get("0001"),
        GlobalCoefficients,
    )
    assert isinstance(
        simai_client._post_processing_directory.get("0002"),
        SurfaceEvolution,
    )
    assert isinstance(simai_client._post_processing_directory.get("0003"), Slice)
    assert isinstance(simai_client._post_processing_directory.get("0004"), VolumeVTU)


def test_post_processing_get_unknown_type(simai_client, httpx_mock):
    """WHEN Calling get() on post-processing directory
    IF an unknown type is received from the server
    THEN a ValueError is raised
    """
    httpx_mock.add_response(
        method="GET",
        url="https://test.test/post-processings/000x",
        json={"id": "000x", "type": "ThisFormatDoesNotExist"},
        status_code=200,
    )

    with pytest.raises(ValueError):
        simai_client._post_processing_directory.get("000x")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.fixture()
def geometries_with_epsilon_values(create_mock_geometry):
    """A list of geometries with a numerical metadata, with values separated
    by steps of 10**-2 and 10**-4, so tolerance can be tested.
    """
    geometries = []
    for length in [
        0.99,
        0.9999,
        1,
        1.0001,
        1.01,
        1.99,
        1.9999,
        2,
        2.0001,
        2.01,
        2.99,
        2.9999,
        3,
        3.0001,
        3.01,
        3.99,
        3.9999,
        4,
        4.0001,
        4.01,
    ]:
        geometries.append(create_mock_geometry(id=f"l{length}", length=length))
    yield geometries


def test_sweep_tolerance_default(geometry_directory, geometries_with_epsilon_values):
    """WHEN Running a sweep with a default tolerance, smaller than the epsilon in metadata values
    THEN every value in the metadata represent a step
      (every increasing order gives 1 more geometry).
    """
    candidate = next(g for g in geometries_with_epsilon_values if g.metadata["length"] == 3)
    assert candidate.metadata["length"] == 3
    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        include_center=False,
    )
    assert {g.metadata["length"] for g in swept} == {2.9999, 3.0001}

    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        order=3,
        include_center=False,
    )
    assert {g.metadata["length"] for g in swept} == {
        2.01,
        2.99,
        2.9999,
        3.0001,
        3.01,
        3.99,
    }


def test_sweep_tolerance_negligible(geometry_directory, geometries_with_epsilon_values):
    """WHEN Running a sweep with a tolerance 10**-5 smaller than the epsilon in values (10**-4)
    THEN every value in the metadata represent a step
        (every increasing order gives 1 more geometry).
    """
    candidate = next(g for g in geometries_with_epsilon_values if g.metadata["length"] == 3)
    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        order=3,
        include_center=False,
        tolerance=10**-5,
    )
    assert {g.metadata["length"] for g in swept} == {
        2.01,
        2.99,
        2.9999,
        3.0001,
        3.01,
        3.99,
    }


def test_sweep_default_tolerance_10_minus_3(geometry_directory, geometries_with_epsilon_values):
    """WHEN Running a sweep with tolerance greater than some diffs in values
    THEN those values are counted together when concerning sweep order (depth)
    """

    tolerance = 10**-3
    candidate = next(g for g in geometries_with_epsilon_values if g.metadata["length"] == 3)
    assert candidate.metadata["length"] == 3
    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        include_center=False,
        tolerance=tolerance,
    )
    # with tolerance 10**-3, 2.9999 is equal to 3, so the 2 neighbors are those:
    expected_without_center = {2.99, 3.01}
    assert {g.metadata["length"] for g in swept} == expected_without_center

    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        include_center=True,
        tolerance=tolerance,
    )
    expected_with_center = expected_without_center.union({2.9999, 3, 3.0001})
    assert {g.metadata["length"] for g in swept} == expected_with_center


def test_sweep_default_tolerance_10_minus_3_order_3(
    geometry_directory, geometries_with_epsilon_values
):
    """WHEN Running a sweep with tolerance greater than some diffs in values
    THEN those values are grouped together when concerning sweep order (depth)
    """
    tolerance = 10**-3
    candidate = next(g for g in geometries_with_epsilon_values if g.metadata["length"] == 3)
    assert candidate.metadata["length"] == 3
    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        include_center=False,
        order=3,
        tolerance=tolerance,
    )
    # With tolerance 10**-3, 2.9999 is equal to 3,
    # Also 1.9999, 2, and 2.0001 are equal so they are counted as 1 step,
    # same with 3.9999, 4, 4.0001.
    expected_without_center = {
        1.9999,
        2,
        2.0001,
        2.01,
        2.99,
        3.01,
        3.99,
        3.9999,
        4,
        4.0001,
    }
    assert {g.metadata["length"] for g in swept} == expected_without_center

    swept = geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_epsilon_values,
        include_center=True,
        order=3,
        tolerance=tolerance,
    )
    expected_with_center = expected_without_center.union({2.9999, 3, 3.0001})
    assert {g.metadata["length"] for g in swept} == expected_with_center



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from ansys.simai.core.api.mixin import ApiClientMixin


class UserClientMixin(ApiClientMixin):
    """Provides the client for the User ("/users") part of the API."""

    def get_offline_tokens(self):
        """List all offline tokens for the current user.

        Returns:
            List of offline token representations.
        """
        return self._get("users/offline-tokens")

    def revoke_offline_token(self, client_id: str) -> None:
        """Revoke an offline token for the current user.

        Args:
            client_id: The client ID of the token to revoke.

        Raises:
            ansys.simai.core.errors.NotFoundError: If no token with that client ID exists.
            ansys.simai.core.errors.ApiClientError: On other HTTP errors.
        """
        self._delete(f"users/offline-tokens/{client_id}", return_json=False)



# Copyright (C) 2023 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# ruff: noqa: INP001, ERA001
#
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
from datetime import datetime
from pathlib import Path

from ansys_sphinx_theme import ansys_favicon, get_version_match, pyansys_logo_black
from sphinx_gallery.sorting import FileNameSortKey

from ansys.simai.core import __version__

# -- Project information -----------------------------------------------------

project = "ansys-simai-core"
author = "ANSYS, Inc."
release = version = __version__
copyright = f"(c) {datetime.now().year} ANSYS, Inc. All rights reserved"
cname = os.getenv("DOCUMENTATION_CNAME", "simai.docs.pyansys.com")

SOURCE_PATH = Path(__file__).parent.resolve().absolute()
ANSYS_SIMAI_THUMBNAIL = str(os.path.join(SOURCE_PATH, "_static", "ansys_simai.png"))

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx_design",
    "sphinxcontrib.autodoc_pydantic",
    "sphinx_gallery.gen_gallery",
    "sphinx_copybutton",
]
# Sphinx Gallery Options

sphinx_gallery_conf = {
    # default png file for thumbnails
    "default_thumb_file": ANSYS_SIMAI_THUMBNAIL,
    # path to your examples scripts
    "examples_dirs": ["examples"],
    # path where to save gallery generated examples
    "gallery_dirs": ["_examples"],
    # Remove the "Download all examples" button from the top level gallery
    "download_all_examples": False,
    # Sort gallery example by file name instead of number of lines (default)
    "within_subsection_order": FileNameSortKey,
}
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pydantic": ("https://docs.pydantic.dev/latest/", None),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Do not display todos in the documentation
todo_include_todos = False

# Autodoc settings
autodoc_typehints = "both"
autodoc_typehints_description_target = "documented"
autodoc_member_order = "groupwise"
autoclass_content = "both"

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Don't always include full modules names to all descriptions
add_module_names = False

# Complain about all broken links
nitpicky = True

nitpick_ignore = [
    ("py:obj", "ansys.simai.core.data.base.DataModelType"),
    ("py:class", "_io.BytesIO"),
    ("py:class", "pydantic_core._pydantic_core.Url"),
    ("py:class", "pydantic_core._pydantic_core.Annotated"),
    ("py:class", "annotated_types.Gt"),
    ("py:class", "pydantic.networks.UrlConstraints"),
    # httpx cannot be linked: https://github.com/encode/httpx/discussions/3091
    ("py:class", "Request"),
]
nitpick_ignore_regex = {
    ("py:class", r"httpx\..+"),
}

source_suffix = ".rst"
master_doc = "index"

# -- Options for HTML output -------------------------------------------------

html_theme = "ansys_sphinx_theme"
html_short_title = html_title = "simai"
html_logo = pyansys_logo_black
html_favicon = ansys_favicon
html_static_path = ["_static"]

suppress_warnings = ["config.cache"]

html_context = {
    "github_user": "ansys",
    "github_repo": "pysimai",
    "github_version": "main",
    "doc_path": "doc",
    "pyansys_tags": ["AI"],
}
html_theme_options = {
    "logo": "pyansys",
    "github_url": "https://github.com/ansys/pysimai",
    "show_prev_next": False,
    "show_breadcrumbs": True,
    "additional_breadcrumbs": [
        ("PyAnsys", "https://docs.pyansys.com/"),
    ],
    "switcher": {
        "json_url": f"https://{cname}/versions.json",
        "version_match": get_version_match(version),
    },
    "show_nav_level": 2,
    "navigation_depth": -1,
    "collapse_navigation": False,
}

# -- Options for LaTeX output -------------------------------

# Remove blank pages
# https://stackoverflow.com/questions/5422997/sphinx-docs-remove-blank-pages-from-generated-pdfs/5741112#5741112

latex_elements = {"extraclassoptions": "openany,oneside"}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_interpolate_geometries:

Interpolating Between Geometries
==============================================

This example demonstrates how to interpolate between two geometries in latent space.

Latent space interpolation allows you to:

- Create smooth transitions between existing designs.
- Explore the design space systematically.
- Generate variations that blend features from multiple geometries.

Before you begin
-------------------------------------------

- Complete ":ref:`ref_build_model`" to train a Generative design model.
- Ensure that the model training has been completed successfully. To do so, verify if a new workspace was created for the trained model.

"""


###############################################################################
# Import necessary libraries
# -------------------------------------------

import os
from typing import List

import ansys.simai.core as asc
from ansys.simai.core.data.geomai.predictions import GeomAIPrediction, GeomAIPredictionConfiguration

###############################################################################
# Configure your settings
# -------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION = "my_organization"  # Replace with your organization name
PROJECT_NAME = "new-bracket-project"  # Replace with your project name
WORKSPACE_NAME = "new-bracket-project #1"  # Typically "{PROJECT_NAME} #{number}"
OUTPUT_DIR = "interpolations"  # Directory to save interpolated geometries
NUM_STEPS = 10  # Number of interpolation steps
RESOLUTION = (100, 100, 100)  # Output resolution (x, y, z)

# Choose geometries to interpolate between (by name)
GEOM_A_NAME = "geometry_name_a"  # Replace with actual geometry name
GEOM_B_NAME = "geometry_name_b"  # Replace with actual geometry name


###############################################################################
# Define the interpolation function
# -------------------------------------------
# Before running predictions, we need a function to interpolate between two latent vectors.


def interpolate_latents(vec1: List[float], vec2: List[float], alpha: float) -> List[float]:
    """Perform linear interpolation between two latent vectors.

    Parameters
    ----------
    vec1 : List[float]
        First latent parameter vector.
    vec2 : List[float]
        Second latent parameter vector.
    alpha : float
        Interpolation factor (0.0 = vec1, 1.0 = vec2).

    Returns
    -------
    List[float]
        Interpolated latent parameter vector.
    """
    return [(1 - alpha) * x + alpha * y for x, y in zip(vec1, vec2)]


###############################################################################
# Initialize the client and get the workspace
# --------------------------------------------------
# Connect to GeomAI and retrieve your trained workspace:

simai_client = asc.SimAIClient(organization=ORGANIZATION)
geomai_client = simai_client.geomai

workspace = geomai_client.workspaces.get(name=WORKSPACE_NAME)
print(f"Using workspace: {workspace.name}")

###############################################################################
# Get latent parameters for all geometries
# -----------------------------------------------
# Download the latent parameters of all training geometries in the workspace:

latent_dict = workspace.get_latent_parameters()

###############################################################################
# Display available geometries
# ----------------------------------
# List all geometries available for interpolation:

print("\nAvailable geometries:")
for i, name in enumerate(latent_dict.keys()):
    print(f"  [{i}] {name}")

###############################################################################
# Validate selected geometries
# -------------------------------------------
# Check if the selected geometries exist:

if GEOM_A_NAME not in latent_dict:
    print(f"Error: Geometry '{GEOM_A_NAME}' not found in workspace.")
    print("Please update GEOM_A_NAME with a valid geometry name from the list above.")
    raise ValueError(f"Geometry '{GEOM_A_NAME}' not found")

if GEOM_B_NAME not in latent_dict:
    print(f"Error: Geometry '{GEOM_B_NAME}' not found in workspace.")
    print("Please update GEOM_B_NAME with a valid geometry name from the list above.")
    raise ValueError(f"Geometry '{GEOM_B_NAME}' not found")

print(f"\nInterpolating from '{GEOM_A_NAME}' to '{GEOM_B_NAME}' in {NUM_STEPS} steps.")

###############################################################################
# Interpolate in latent space and generate geometries
# ----------------------------------------------------------
# Generate intermediate geometries by linearly interpolating in latent space:

vec_a = latent_dict[GEOM_A_NAME]
vec_b = latent_dict[GEOM_B_NAME]

predictions: list[GeomAIPrediction] = []

for i in range(NUM_STEPS + 1):
    alpha = i / NUM_STEPS
    latent_params = interpolate_latents(vec_a, vec_b, alpha)
    print(f"\nGenerating geometry {i}/{NUM_STEPS} (alpha={alpha:.2f})...")
    config = GeomAIPredictionConfiguration(
        latent_params=latent_params,
        resolution=RESOLUTION,
    )
    prediction = geomai_client.predictions.run(config, workspace)
    print(f"Prediction {i}: {prediction.id} started...")
    predictions.append(prediction)


###############################################################################
# Download generated geometries
# -------------------------------------------

for i, prediction in enumerate(predictions):
    if prediction.wait(timeout=600):  # Wait up to 10 minutes
        if prediction.has_failed:
            print(f"✗ Prediction {i} failed: {prediction.failure_reason}")
            continue

        # Save result
        out_dir = os.path.join(OUTPUT_DIR, workspace.name)
        os.makedirs(out_dir, exist_ok=True)
        out_path = os.path.join(out_dir, f"prediction_{i:02d}_{prediction.id}.vtp")
        geomai_client.predictions.download(prediction.id, out_path)
        print(f"✓ Saved prediction to {out_path}")
    else:
        print(f"✗ Prediction {i} timed out.")

###############################################################################
# The downloaded VTP files can be used for:
#
# - Visualization in your usual solver.
# - SimAI training data or predictions.
# - Further analysis and post-processing.

###############################################################################
# Next steps
# -------------------------------------------
# To go further, you can:
#
# - Interpolate between more than two geometries.
# - Combine interpolation with optimization for specific design goals.



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
.. _ref_list_based_subset_assignment:

List-based subset assignment
============================

This example demonstrates how to distribute your dataset between Test
and Training subsets using lists.

"""

###############################################################################
# Import necessary libraries
# --------------------------

import ansys.simai.core as asc

###############################################################################
# Create lists
# ------------
# List the data to be used for the test set.

# CAUTION:
# All training data that are not included into the following list will be
# assigned the training subset
TEST_LIST = ["My_td_1", "My_td_2"]

###############################################################################
# Connect to the platform
# ------------------------
# Connect to the SimAI platform. Refer to the :ref:`anchor-credentials`
# section of the documentation to adapt the connection type.

simai_client = asc.SimAIClient(organization="My_organization_name")
project = simai_client.projects.get(name="My_project_name")
###############################################################################
# Assign subsets
# --------------
# Assign a subset to each dataset (list) you created.

td_list = project.data
for td in td_list:
    if td.name in TEST_LIST:
        td.assign_subset(project, "Test")
    else:
        td.assign_subset(project, "Training")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest


@pytest.fixture()
def geometries_with_one_metadata(create_mock_geometry):
    """A list of geometries with:
    - 1 numerical metadata (length), range 0 to 9
    Used to test basic sweep on 1 dimension.
    """
    geometries = []
    for i in range(10):
        geometries.append(create_mock_geometry(id=f"l-{i}", length=i))
    yield geometries


@pytest.fixture()
def geometries_with_one_usable_metadata(create_mock_geometry):
    """A list of geometries with:
    - 1 numerical metadata (length), range 0 to 9,
    - 1 string metadata (type)
    Used to test conditions when string metadata is included or ignored
    """
    geometries = []
    for i in range(10):
        geometries.append(create_mock_geometry(id=f"l-{i}", length=i, type=str(i)))
    yield geometries


@pytest.fixture()
def geometries_with_no_usable_metadata(create_mock_geometry):
    """A list of geometries with no numerical metadata (only a string one)
    Used to test failure case
    """
    geometries = []
    for i in range(10):
        geometries.append(create_mock_geometry(id=f"l-{i}", type=str(i)))
    yield geometries


def test_sweep_one_dimension_gives_two_neighbors(geometry_directory, geometries_with_one_metadata):
    """WHEN Running a sweep in the middle of a single dimension
    THEN two neighbors are returned
    ALSO include_diagonals has no effect with 1 metadata
    """
    candidate = geometries_with_one_metadata[5]
    assert candidate.metadata["length"] == 5
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {4, 6}


def test_sweep_one_dimension_with_include_center_gives_three_results(
    geometry_directory, geometries_with_one_metadata
):
    """WHEN Running a sweep in the middle of a single dimension with include_center
    THEN three neighbors are returned
    ALSO include_diagonals has no effect with 1 metadata
    """
    candidate = geometries_with_one_metadata[5]
    assert candidate.metadata["length"] == 5
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            include_center=True,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {4, 5, 6}


def test_sweep_one_dimension_along_explicit_variable(
    geometry_directory, geometries_with_one_metadata
):
    """WHEN Running a sweep along explicit variable
    THEN two neighbors are returned
    ALSO include_diagonals has no effect with 1 metadata
    """
    candidate = geometries_with_one_metadata[5]
    assert candidate.metadata["length"] == 5
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            swept_metadata="length",
            geometries=geometries_with_one_metadata,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {4, 6}

        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            swept_metadata=["length"],
            geometries=geometries_with_one_metadata,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {4, 6}


def test_sweep_one_dimension_with_order(geometry_directory, geometries_with_one_metadata):
    """WHEN Running a sweep in the middle of a single dimension with an order parameter
    THEN 2 * order neighbors are returned if possible
    ALSO include_diagonals has no effect with 1 metadata
    """
    candidate = geometries_with_one_metadata[5]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=2,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {3, 4, 6, 7}

    candidate = geometries_with_one_metadata[1]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=3,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {0, 2, 3, 4}

    candidate = geometries_with_one_metadata[0]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=3,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {1, 2, 3}


def test_sweep_one_dimension_with_order_and_include_center(
    geometry_directory, geometries_with_one_metadata
):
    """WHEN Running a sweep in the middle of a single dimension with order and include_center
    THEN 2 * order + 1 neighbors are returned when possible
    ALSO include_diagonals has no effect with 1 metadata
    """
    candidate = geometries_with_one_metadata[5]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=2,
            include_center=True,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {3, 4, 5, 6, 7}

    candidate = geometries_with_one_metadata[1]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=3,
            include_center=True,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {0, 1, 2, 3, 4}

    candidate = geometries_with_one_metadata[0]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            order=3,
            include_center=True,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {0, 1, 2, 3}


def test_sweep_one_dimension_at_min_gives_one_neighbor(
    geometry_directory, geometries_with_one_metadata
):
    """WHEN Running a sweep at the min border of a single dimension
    THEN only one neighbor is returned
    """
    candidate = geometries_with_one_metadata[0]
    assert candidate.metadata["length"] == 0
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            include_diagonals=include_diagonals,
        )
        assert len(swept) == 1
        assert swept[0].metadata["length"] == 1


def test_sweep_one_dimension_at_max_gives_one_neighbor(
    geometry_directory, geometries_with_one_metadata
):
    """WHEN Running a sweep at the max border of a single dimension
    THEN only one neighbor is returned
    """
    candidate = geometries_with_one_metadata[9]
    assert candidate.metadata["length"] == 9
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            include_diagonals=include_diagonals,
        )
        assert len(swept) == 1
        assert swept[0].metadata["length"] == 8


def test_sweep_all_columns_text_metadata_ignored(
    geometry_directory, geometries_with_one_usable_metadata
):
    """WHEN Running a sweep without specifying columns
    THEN textual metadata is ignored, and sweep only uses numerical metadata
    """
    candidate = geometries_with_one_usable_metadata[1]
    for include_diagonals in [False, True]:
        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_usable_metadata,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {0, 2}

        swept = geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_usable_metadata,
            include_center=True,
            include_diagonals=include_diagonals,
        )
        assert {g.metadata["length"] for g in swept} == {0, 1, 2}


def test_sweep_unknown_metadata_column_raises(
    geometry_directory, geometries_with_one_usable_metadata
):
    """WHEN Running a sweep with specifying an unknown metadata column
    THEN a ValueError is raised
    """
    candidate = geometries_with_one_usable_metadata[4]
    for include_diagonals in [False, True]:
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["length", "not-existing"],
                geometries=geometries_with_one_usable_metadata,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["toto"],
                geometries=geometries_with_one_usable_metadata,
                include_diagonals=include_diagonals,
            )


def test_sweep_with_text_metadata_fails(geometry_directory, geometries_with_one_usable_metadata):
    """WHEN Running a sweep using a text metadata
    THEN a ValueError is raised
    """
    candidate = geometries_with_one_usable_metadata[4]
    for include_diagonals in [False, True]:
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["length", "type"],
                geometries=geometries_with_one_usable_metadata,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["type"],
                geometries=geometries_with_one_usable_metadata,
                include_diagonals=include_diagonals,
            )


def test_sweep_with_no_metadata_fails(geometry_directory, geometries_with_no_usable_metadata):
    """WHEN Running a sweep using no metadata column
    THEN a ValueError is raised
    """
    candidate = geometries_with_no_usable_metadata[4]
    for include_diagonals in [False, True]:
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_no_usable_metadata,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=["type"],
                geometries=geometries_with_no_usable_metadata,
                include_diagonals=include_diagonals,
            )


def test_sweep_incorrect_metadata_argument(geometry_directory, geometries_with_one_metadata):
    """WHEN Running a sweep passing a metadata argument neither List[str] nor str,
    THEN a TypeError is raised
    """
    candidate = geometries_with_one_metadata[4]
    for include_diagonals in [False, True]:
        with pytest.raises(TypeError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=6,
                geometries=geometries_with_one_metadata,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(TypeError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=[6],
                geometries=geometries_with_one_metadata,
                include_diagonals=include_diagonals,
            )


def test_sweep_incorrect_order_argument(geometry_directory, geometries_with_one_metadata):
    """WHEN Running a sweep passing an order argument not int
    THEN a TypeError is raised
    WHEN Running a sweep passing an order argument int < 1
    THEN a ValueError is raised
    """
    candidate = geometries_with_one_metadata[4]
    for include_diagonals in [False, True]:
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_one_metadata,
                order=0,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(ValueError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=[6],
                geometries=geometries_with_one_metadata,
                order=-3,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(TypeError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                geometries=geometries_with_one_metadata,
                order=2.4,
                include_diagonals=include_diagonals,
            )
        with pytest.raises(TypeError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=[6],
                geometries=geometries_with_one_metadata,
                order="blabla",
                include_diagonals=include_diagonals,
            )
        with pytest.raises(TypeError):
            geometry_directory.sweep(
                candidate_geometry=candidate,
                swept_metadata=[6],
                geometries=geometries_with_one_metadata,
                order=[4],
                include_diagonals=include_diagonals,
            )


def test_sweep_incorrect_tolerance_argument(geometry_directory, geometries_with_one_metadata):
    """WHEN Running a tolerance passing an order argument not number
    THEN a TypeError is raised
    WHEN Running a sweep passing an tolerance argument < 0
    THEN a ValueError is raised
    """
    candidate = geometries_with_one_metadata[4]
    geometry_directory.sweep(
        candidate_geometry=candidate,
        geometries=geometries_with_one_metadata,
        tolerance=0,
    )
    with pytest.raises(ValueError):
        geometry_directory.sweep(
            candidate_geometry=candidate,
            swept_metadata=[6],
            geometries=geometries_with_one_metadata,
            tolerance=-3,
        )
    with pytest.raises(TypeError):
        geometry_directory.sweep(
            candidate_geometry=candidate,
            geometries=geometries_with_one_metadata,
            tolerance="zzz",
        )
    with pytest.raises(TypeError):
        geometry_directory.sweep(
            candidate_geometry=candidate,
            swept_metadata=[6],
            geometries=geometries_with_one_metadata,
            tolerance=[4],
        )



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_build_model:

Building a Generative Design Model
================================================

This example demonstrates how to configure a Generative Design model, start the model training process, and monitor the build progress.

Before you begin
-------------------------------------------

- Complete ":ref:`ref_create_project_upload_data`" to create a project with training data.
- Ensure all training data in your project are ready (processed successfully).

"""

###############################################################################
# Import necessary libraries
# -------------------------------------------

import ansys.simai.core as asc
from ansys.simai.core.data.geomai.models import GeomAIModelConfiguration

###############################################################################
# Configure your settings
# -------------------------------------------
# Update these variables with your specific settings:

ORGANIZATION = "my_organization"  # Replace with your organization name
PROJECT_NAME = "new-bracket-project"  # Replace with your project name
BUILD_PRESET = "default"  # Options: "debug", "short", "default", "long"

###############################################################################
# The ``BUILD_PRESET`` options correspond to:
#
# - ``"debug"``: Fast training for testing (very few epochs).
# - ``"short"``: Quick training with reduced accuracy.
# - ``"default"``: Balanced training time and quality.
# - ``"long"``: Longer training for best quality.


###############################################################################
# Initialize the client and get the project
# -------------------------------------------
# Connect to the instance:

simai_client = asc.SimAIClient(organization=ORGANIZATION)
geomai_client = simai_client.geomai

###############################################################################
# Retrieve the project by name:

project = geomai_client.projects.get(name=PROJECT_NAME)
print(f"Using project: {project.name}")

###############################################################################
# Verify project data are ready
# -------------------------------------------
# Before building a model, ensure all training data are processed:

project_data = project.data()
ready_data = [data for data in project_data if data.is_ready]
print(f"Project has {len(ready_data)}/{len(project_data)} ready training data")

if len(ready_data) < len(project_data):
    print("Warning: Some training data are not ready. Model may fail to build.")

###############################################################################
# Configure the model
# -------------------------------------------
# To define the configuration for your model, you can either specify the build preset or the number
# of epochs.
# To do so, instead of ``build_preset``, you can specify the number of epochs directly. Example: ``nb_epochs=100``.
#
# The number of latent parameters defines the complexity of the model's latent space; start with a small number (e.g., 10) and adjust based on your needs.


configuration = GeomAIModelConfiguration(
    build_preset=BUILD_PRESET,
)

###############################################################################
# Build the model
# -------------------------------------------
# Start the model training process:

model = geomai_client.models.build(project, configuration)
print(f"Started build for model {model.id}")

###############################################################################
# Wait for model training to complete
# -------------------------------------------
# Monitor the training process and handle completion:

print("Waiting for model training to complete (this may take a while)...")
while not model.wait(timeout=60):  # Check every 60 seconds
    print(f"Model {model.id} is still training...")

if model.has_failed:
    print(f"Model training failed: {model.failure_reason}")
else:
    print(f"Model {model.id} trained successfully and is ready to use.")

###############################################################################
# Next steps
# -------------------------------------------
# Once your model is trained, you can:
#
# - Generate random geometries: :ref:`ref_generate_random_geometries`
# - Interpolate between geometries: :ref:`ref_interpolate_geometries`



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _ref_basic_create_project_upload_data:

Creating a SimAI Project and Uploading Training Data
=======================================================

This example demonstrates how to connect to SimAI, create a new project, and upload training data folders.

Before you begin
-------------------

Make sure you have:

- Valid SimAI credentials and organization access.
- A dataset folder containing subdirectories with your training data.
- The ``ansys-simai-core`` library installed.

"""

###############################################################################
# Import necessary libraries
# ----------------------------------

import os

import ansys.simai.core as asc
from ansys.simai.core.data.training_data import TrainingData
from ansys.simai.core.errors import NotFoundError

###############################################################################
# Configure your settings
# ----------------------------------
# Update these variables with your specific settings:

ORGANIZATION_NAME = "<your_organization>"  # Replace with your organization name
PROJECT_NAME = "<your_project_name>"  # Your project name
DATASET_PATH = "<PATH_TO_YOUR_DATASET>"  # Directory containing subdirectories with training data

###############################################################################
# Initialize the SimAI client
# ----------------------------------
# Create a client to connect to the SimAI platform:

simai_client = asc.SimAIClient(organization=ORGANIZATION_NAME)

###############################################################################
# Set up the project
# ----------------------------------
# Try to get an existing project by name, or create it if it does not exist:

try:
    project = simai_client.projects.get(name=PROJECT_NAME)
    print(f"Using existing project: {PROJECT_NAME}")
except NotFoundError:
    project = simai_client.projects.create(PROJECT_NAME)
    print(f"Created new project: {PROJECT_NAME}")

###############################################################################
# Set as the current working project
# ----------------------------------
# Setting the current project allows subsequent operations to use this project by default:

simai_client.set_current_project(PROJECT_NAME)
print(f"Current project: {simai_client.current_project}")

###############################################################################
# Upload training data
# ----------------------------------
# Upload all directories from the dataset path as training data.
# Each subdirectory should contain the files for one training data sample.


available_tds = simai_client.training_data.list()

print("\nUploading training data files:")
successful_uploads = 0
failed_uploads = 0

for dir in os.listdir(DATASET_PATH):
    complete_path = os.path.join(DATASET_PATH, dir)
    print(f"Uploading {dir}")

    existing_tds = [td for td in available_tds if td.name == dir]
    if existing_tds:
        print(f"Training data '{dir}' already exists in the datalake. Skipping upload.")
        td = existing_tds[0]
        try:
            td.add_to_project(project)
            print(f"✓ Added existing '{dir}' to project '{project.name}'")
            successful_uploads += 1
        except Exception as e:
            print(f"✗ Failed to add existing '{dir}' to project: {e}")
            failed_uploads += 1
        continue
    else:
        try:
            td: TrainingData = simai_client.training_data.create(dir)
            td.upload_folder(complete_path)
            print(f"Uploaded {dir} successfully.")
        except Exception as e:
            print(f"Failed to upload {dir}: {e}")
            failed_uploads += 1
            continue


print(f"\nUpload summary: {successful_uploads} successful, {failed_uploads} failed")

###############################################################################
# Check and wait for data processing
# -------------------------------------
# After uploading, SimAI needs to process the training data.
# Get all data in the current project and wait for them to be ready.

project_data = project.data

print("\nWaiting for data processing to complete...")
for data in project_data:
    print(f"Processing {data.name}...")
    data.wait()
    print(f"Data '{data.name}' is ready")

###############################################################################
# Display project status summary
# -------------------------------------
# Show a summary of the project's data processing status:

print("\nProject Summary")
print("=" * 50)

ready_data = [data for data in project_data if data.is_ready]
not_ready_data = [data for data in project_data if not data.is_ready]

print(f"Total data in project: {len(project_data)}")
print(f"Ready data: {len(ready_data)} of {len(project_data)}")
print(f"Not ready data: {len(not_ready_data)} of {len(project_data)}")

if not_ready_data:
    print("\nFailed data processing details:")
    for data in not_ready_data:
        print(f"- {data.name}: {data.failure_reason}")

###############################################################################
# Next steps
# ----------------------------------
# Once all data are ready, you can proceed to configure and build a model.
# See the next tutorial: :ref:`ref_basic_build_model`



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pathlib
from io import BytesIO

import pytest

from ansys.simai.core.data.base import DataModel, Directory
from ansys.simai.core.data.types import (
    Range,
    are_scalars_equal,
    build_scalars,
    get_id_from_identifiable,
    get_object_from_identifiable,
    is_scalars,
    to_raw_filters,
    unpack_named_file,
)


def test_build_scalars(mocker):
    is_bc = mocker.patch("ansys.simai.core.data.types.is_scalars")
    is_bc.return_value = True
    assert build_scalars({"Vx": 10.1}) == {"Vx": 10.1}
    assert build_scalars({"Vx": 10, "Vy": 2}) == {"Vx": 10, "Vy": 2}
    assert build_scalars({"Vx": 10}, Vy=2) == {"Vx": 10, "Vy": 2}
    assert build_scalars(Vx=10, Vy=2) == {"Vx": 10, "Vy": 2}
    assert build_scalars() == {}
    assert build_scalars(scalars={}) == {}

    is_bc.return_value = False
    with pytest.raises(ValueError):
        build_scalars({"Vx": 10})


def test_basic_range():
    epsilon = 10**-7
    a_range = Range(10, 30)
    assert not a_range.match_value(-(10**8))
    assert not a_range.match_value(-30)
    assert not a_range.match_value(0)
    assert not a_range.match_value(10**-6)

    assert not a_range.match_value(a_range.min - 20 * epsilon)
    assert a_range.match_value(a_range.min - epsilon)
    assert a_range.match_value(a_range.min)
    assert a_range.match_value(a_range.min + epsilon)
    assert a_range.match_value(a_range.min + 20 * epsilon)

    assert a_range.match_value(20)

    assert a_range.match_value(a_range.max - 20 * epsilon)
    assert a_range.match_value(a_range.max - epsilon)
    assert a_range.match_value(a_range.max)
    assert a_range.match_value(a_range.max + epsilon)
    assert not a_range.match_value(a_range.max + 20 * epsilon)

    assert not a_range.match_value(40)
    assert not a_range.match_value(10**8)
    assert not a_range.match_value("toto")


def test_min_only_range():
    epsilon = 10**-7
    a_range = Range(min=10)
    assert not a_range.match_value(-(10**8))
    assert not a_range.match_value(-30)
    assert not a_range.match_value(0)
    assert not a_range.match_value(10**-6)

    assert not a_range.match_value(a_range.min - 20 * epsilon)
    assert a_range.match_value(a_range.min - epsilon)
    assert a_range.match_value(a_range.min)
    assert a_range.match_value(a_range.min + epsilon)
    assert a_range.match_value(a_range.min + 20 * epsilon)

    assert a_range.match_value(20)
    assert a_range.match_value(30)
    assert a_range.match_value(40)
    assert a_range.match_value(10**8)


def test_max_only_range():
    epsilon = 10**-7
    a_range = Range(max=30)
    assert a_range.match_value(-(10**8))
    assert a_range.match_value(-30)
    assert a_range.match_value(0)
    assert a_range.match_value(10**-6)
    assert a_range.match_value(10)
    assert a_range.match_value(20)

    assert a_range.match_value(a_range.max - 20 * epsilon)
    assert a_range.match_value(a_range.max - epsilon)
    assert a_range.match_value(a_range.max)
    assert a_range.match_value(a_range.max + epsilon)
    assert not a_range.match_value(a_range.max + 20 * epsilon)

    assert not a_range.match_value(40)
    assert not a_range.match_value(10**8)


def test_scalar_identity():
    assert is_scalars({"Vx": 10})
    assert is_scalars({"Vx": 10, "Vy": 0, "Vz": 0})
    assert not is_scalars({"Vx": "beep"})
    assert not is_scalars({"Vx": 10, "Vy": "boop"})
    assert not is_scalars([0])
    assert not is_scalars((0))


def test_scalar_equality():
    assert are_scalars_equal({"Vx": 10}, {"Vx": 10})
    assert are_scalars_equal({"Vx": 10, "Vy": 0}, {"Vx": 10, "Vy": 0})
    assert are_scalars_equal({"Vy": 0, "Vx": 10}, {"Vx": 10, "Vy": 0})
    assert are_scalars_equal({"Vx": 10**-7}, {"Vx": 0})
    assert not are_scalars_equal({"Vx": 10**-7}, {"Vx": 0}, tolerance=10**-8)


def test_scalar_equality_wrong_parameters():
    with pytest.raises(TypeError):
        are_scalars_equal({"Vx": 10}, "toto")
    with pytest.raises(TypeError):
        are_scalars_equal("buz", {"Vx": 10})
    with pytest.raises(ValueError):
        are_scalars_equal({"Vx": 10}, {"Vx": 10}, tolerance=-5)


def test_unpack_named_file(mocker):
    open_file = mocker.patch("ansys.simai.core.data.types.file_path_to_obj_file")
    fake_file = mocker.MagicMock()
    open_file.return_value = fake_file

    with unpack_named_file("test.stl") as (file, name, extension):
        assert file == fake_file
        assert name == "test"
        assert extension == "stl"
        open_file.assert_called_once()
    fake_file.close.assert_called_once()
    open_file.reset_mock()
    fake_file.reset_mock()

    with unpack_named_file(pathlib.Path("test.stl")) as (file, name, extension):
        assert file == fake_file
        assert name == "test"
        assert extension == "stl"
        open_file.assert_called_once()
    fake_file.close.assert_called_once()
    open_file.reset_mock()
    fake_file.reset_mock()

    with unpack_named_file(("test.stl", "override.cgns")) as (file, name, extension):
        assert file == fake_file
        assert name == "override"
        assert extension == "cgns"
        open_file.assert_called_once()
    fake_file.close.assert_called_once()
    open_file.reset_mock()
    fake_file.reset_mock()

    binary_file_obj = BytesIO(b"blah")
    with unpack_named_file((binary_file_obj, "test.stl")) as (file, name, extension):
        assert file == binary_file_obj
        assert name == "test"
        assert extension == "stl"
        open_file.assert_not_called()
    fake_file.close.assert_not_called()
    open_file.reset_mock()
    fake_file.reset_mock()

    with unpack_named_file(("f.zip.stl", "¹³³⁷/ᵤₛₑᵣ-aₚᵢ\\n-..cgns")) as (
        file,
        name,
        extension,
    ):
        assert file == fake_file
        assert name == "¹³³⁷/ᵤₛₑᵣ-aₚᵢ\\n-."
        assert extension == "cgns"
        open_file.assert_called_once()
    fake_file.close.assert_called_once()


def test_get_id_from_identifiable():
    class MyTestModel(DataModel):
        pass

    test_model = MyTestModel(None, None, {"id": "yo"})

    assert get_id_from_identifiable(test_model) == "yo"
    assert get_id_from_identifiable("yo") == "yo"


def test_get_object_from_identifiable():
    class MyTestModel(DataModel):
        pass

    test_model = MyTestModel(None, None, {"id": "yo"})

    class MyTestDirectory(Directory):
        _data_model = MyTestModel

        def get(self, id: str):
            if id == "yo":
                return test_model
            else:
                raise ValueError

    directory = MyTestDirectory(client=None)

    assert get_object_from_identifiable(test_model, directory) == test_model
    assert get_object_from_identifiable("yo", directory) == test_model


@pytest.mark.parametrize(
    ("filters", "expected_raw_filters"),
    [
        (
            {"name": "paul", "production_capacity": "6e34Kg"},
            [
                {"field": "name", "operator": "EQ", "value": "paul"},
                {"field": "production_capacity", "operator": "EQ", "value": "6e34Kg"},
            ],
        ),
        (
            [("name", "EQ", "paul"), ("production_capacity", "GTE", 10**10)],
            [
                {"field": "name", "operator": "EQ", "value": "paul"},
                {"field": "production_capacity", "operator": "GTE", "value": 10**10},
            ],
        ),
        (
            [{"field": "nofield", "operator": "GTE", "value": 9000}],
            [{"field": "nofield", "operator": "GTE", "value": 9000}],
        ),
    ],
    ids=["simple_filters", "complex_filters", "raw_filters"],
)
def test_to_raw_filters(filters, expected_raw_filters):
    filters = to_raw_filters(filters)
    assert filters == expected_raw_filters
